@@ -33,34 +33,34 @@
  * limitations under the License.
  */
 
-//! A crate for consuming the runc binary in your Rust applications, similar to [go-runc](https://github.com/containerd/go-runc) for Go.
+//! A crate for consuming the runc binary in your Rust applications, similar to
+//! [go-runc](https://github.com/containerd/go-runc) for Go.
+use std::fmt::{self, Debug, Display};
+use std::path::{Path, PathBuf};
+use std::process::{ExitStatus, Stdio};
+use std::sync::Arc;
 
-use std::fmt::{self, Display};
-use std::path::Path;
-use std::process::{ExitStatus, Output, Stdio};
-use std::time::Duration;
+#[cfg(feature = "async")]
+use async_trait::async_trait;
+#[cfg(feature = "async")]
+use log::debug;
+use oci_spec::runtime::{LinuxResources, Process};
 
-use oci_spec::runtime::{Linux, Process};
+use crate::container::Container;
+use crate::error::Error;
+use crate::options::*;
+use crate::utils::write_value_to_temp_file;
 
-// suspended for difficulties
-// pub mod console;
 pub mod container;
 pub mod error;
 pub mod events;
 pub mod io;
+#[cfg(feature = "async")]
 pub mod monitor;
 pub mod options;
-mod runc;
-mod utils;
+pub mod utils;
 
-use crate::container::Container;
-use crate::error::Error;
-use crate::events::{Event, Stats};
-use crate::monitor::{DefaultMonitor, Exit, ProcessMonitor};
-use crate::options::*;
-use crate::utils::{JSON, TEXT};
-
-type Result<T> = std::result::Result<T, crate::error::Error>;
+pub type Result<T> = std::result::Result<T, crate::error::Error>;
 
 /// Response is for (pid, exit status, outputs).
 #[derive(Debug, Clone)]
@@ -92,163 +92,57 @@ impl Display for LogFormat {
     }
 }
 
-/// Configuration for runc client.
-///
-/// This struct provide chaining interface like, for example, [`std::fs::OpenOptions`].
-/// Note that you cannot access the members of Config directly.
-///
-/// # Example
-///
-/// ```ignore
-/// use runc::{LogFormat, Config};
-///
-/// let config = Config::new()
-///     .root("./new_root")
-///     .debug(false)
-///     .log("/path/to/logfile.json")
-///     .log_format(LogFormat::Json)
-///     .rootless(true);
-/// let client = config.build();
-/// ```
-#[derive(Debug, Clone, Default)]
-pub struct Config(runc::Config);
-
-impl Config {
-    pub fn new() -> Self {
-        Self::default()
-    }
-
-    pub fn command<P>(&mut self, command: P) -> &mut Self
-    where
-        P: AsRef<Path>,
-    {
-        self.0.command(command);
-        self
-    }
-
-    pub fn root<P>(&mut self, root: P) -> &mut Self
-    where
-        P: AsRef<Path>,
-    {
-        self.0.root(root);
-        self
-    }
-
-    pub fn debug(&mut self, debug: bool) -> &mut Self {
-        self.0.debug(debug);
-        self
-    }
-
-    pub fn log<P>(&mut self, log: P) -> &mut Self
-    where
-        P: AsRef<Path>,
-    {
-        self.0.log(log);
-        self
-    }
-
-    pub fn log_format(&mut self, log_format: LogFormat) -> &mut Self {
-        self.0.log_format(log_format);
-        self
-    }
-
-    pub fn log_format_json(&mut self) -> &mut Self {
-        self.0.log_format_json();
-        self
-    }
-
-    pub fn log_format_text(&mut self) -> &mut Self {
-        self.0.log_format_text();
-        self
-    }
-
-    pub fn systemd_cgroup(&mut self, systemd_cgroup: bool) -> &mut Self {
-        self.0.systemd_cgroup(systemd_cgroup);
-        self
-    }
-
-    // FIXME: criu is not supported now
-    // pub fn criu(mut self, criu: bool) -> Self {
-    //     self.0.criu(criu);
-    //     self
-    // }
-
-    pub fn rootless(&mut self, rootless: bool) -> &mut Self {
-        self.0.rootless(rootless);
-        self
-    }
-
-    pub fn set_pgid(&mut self, set_pgid: bool) -> &mut Self {
-        self.0.set_pgid(set_pgid);
-        self
-    }
-
-    pub fn rootless_auto(&mut self) -> &mut Self {
-        self.0.rootless_auto();
-        self
-    }
-
-    pub fn timeout(&mut self, millis: u64) -> &mut Self {
-        self.0.timeout(millis);
-        self
-    }
-
-    pub fn build(&mut self) -> Result<Client> {
-        Ok(Client(self.0.build()?))
-    }
-
-    pub fn build_async(&mut self) -> Result<AsyncClient> {
-        Ok(AsyncClient(self.0.build()?))
+impl Default for LogFormat {
+    fn default() -> Self {
+        LogFormat::Text
     }
 }
 
+#[cfg(not(feature = "async"))]
+pub type Command = std::process::Command;
+
+#[cfg(feature = "async")]
+pub type Command = tokio::process::Command;
+
 #[derive(Debug, Clone)]
-pub struct Client(runc::Runc);
+pub struct Runc {
+    command: PathBuf,
+    args: Vec<String>,
+    spawner: Arc<dyn Spawner + Send + Sync>,
+}
 
-impl Client {
-    /// Create a new runc client from the supplied configuration
-    pub fn from_config(mut config: Config) -> Result<Self> {
-        config.build()
-    }
+impl Runc {
+    fn command(&self, args: &[String]) -> Result<Command> {
+        let args = [&self.args, args].concat();
+        let mut cmd = Command::new(&self.command);
+
+        // Default to piped stdio, and they may be override by command options.
+        cmd.stdin(Stdio::null())
+            .stdout(Stdio::piped())
+            .stderr(Stdio::piped());
+
+        // NOTIFY_SOCKET introduces a special behavior in runc but should only be set if invoked from systemd
+        cmd.args(&args).env_remove("NOTIFY_SOCKET");
 
-    #[cfg(target_os = "linux")]
-    pub fn command(&self, args: &[String]) -> Result<std::process::Command> {
-        let args = [&self.0.args()?, args].concat();
-        let mut cmd = std::process::Command::new(&self.0.command);
-        cmd.args(&args).env_remove("NOTIFY_SOCKET"); // NOTIFY_SOCKET introduces a special behavior in runc but should only be set if invoked from systemd
         Ok(cmd)
     }
+}
 
-    #[cfg(not(target_os = "linux"))]
-    pub fn command(&self, _args: &[String]) -> Result<std::process::Command> {
-        Err(Error::Unimplemented("command".to_string()))
-    }
-
-    pub fn checkpoint(&self) -> Result<()> {
-        Err(Error::Unimplemented("checkpoint".to_string()))
-    }
-
-    fn launch(&self, mut cmd: std::process::Command, combined_output: bool) -> Result<Response> {
-        let child = cmd.spawn().map_err(Error::ProcessSpawnFailed)?;
-        let pid = child.id();
-        let result = child.wait_with_output().map_err(Error::InvalidCommand)?;
-        let status = result.status;
-        let stdout = String::from_utf8(result.stdout).unwrap();
-        let stderr = String::from_utf8(result.stderr).unwrap();
+#[cfg(not(feature = "async"))]
+impl Runc {
+    fn launch(&self, cmd: Command, combined_output: bool) -> Result<Response> {
+        let (status, pid, stdout, stderr) = self.spawner.execute(cmd)?;
         if status.success() {
-            if combined_output {
-                Ok(Response {
-                    pid,
-                    status,
-                    output: stdout + stderr.as_str(),
-                })
+            let output = if combined_output {
+                stdout + stderr.as_str()
             } else {
-                Ok(Response {
-                    pid,
-                    status,
-                    output: stdout,
-                })
-            }
+                stdout
+            };
+            Ok(Response {
+                pid,
+                status,
+                output,
+            })
         } else {
             Err(Error::CommandFailed {
                 status,
@@ -274,10 +168,10 @@ impl Client {
         args.push(id.to_string());
         let mut cmd = self.command(&args)?;
         match opts {
-            Some(CreateOpts { io: Some(_io), .. }) => {
-                _io.set(&mut cmd).map_err(Error::UnavailableIO)?;
+            Some(CreateOpts { io: Some(io), .. }) => {
+                io.set(&mut cmd).map_err(|e| Error::IoSet(e.to_string()))?;
                 let res = self.launch(cmd, true)?;
-                _io.close_after_start();
+                io.close_after_start();
                 Ok(res)
             }
             _ => self.launch(cmd, true),
@@ -297,19 +191,23 @@ impl Client {
 
     /// Execute an additional process inside the container
     pub fn exec(&self, id: &str, spec: &Process, opts: Option<&ExecOpts>) -> Result<()> {
-        let filename = utils::temp_filename_in_runtime_dir()?;
-        let spec_json = serde_json::to_string(spec).map_err(Error::JsonDeserializationFailed)?;
-        std::fs::write(&filename, spec_json).map_err(Error::SpecFileCreationFailed)?;
-        let mut args = vec!["exec".to_string(), "process".to_string(), filename];
+        let (_temp_file, filename) = write_value_to_temp_file(spec)?;
+        let mut args = vec!["exec".to_string(), "--process".to_string(), filename];
         if let Some(opts) = opts {
             args.append(&mut opts.args()?);
         }
         args.push(id.to_string());
         let mut cmd = self.command(&args)?;
-        if let Some(ExecOpts { io: Some(_io), .. }) = opts {
-            _io.set(&mut cmd).map_err(Error::UnavailableIO)?;
+        match opts {
+            Some(ExecOpts { io: Some(io), .. }) => {
+                io.set(&mut cmd).map_err(|e| Error::IoSet(e.to_string()))?;
+                self.launch(cmd, true)?;
+                io.close_after_start();
+            }
+            _ => {
+                self.launch(cmd, true)?;
+            }
         }
-        let _ = self.launch(cmd, true)?;
         Ok(())
     }
 
@@ -327,9 +225,10 @@ impl Client {
 
     /// List all containers associated with this runc instance
     pub fn list(&self) -> Result<Vec<Container>> {
-        let args = ["list".to_string(), "--format-json".to_string()];
+        let args = ["list".to_string(), "--format=json".to_string()];
         let res = self.launch(self.command(&args)?, true)?;
         let output = res.output.trim();
+
         // Ugly hack to work around golang
         Ok(if output == "null" {
             Vec::new()
@@ -345,15 +244,37 @@ impl Client {
         Ok(())
     }
 
+    /// Resume a container
+    pub fn resume(&self, id: &str) -> Result<()> {
+        let args = ["resume".to_string(), id.to_string()];
+        let _ = self.launch(self.command(&args)?, true)?;
+        Ok(())
+    }
+
+    pub fn checkpoint(&self) -> Result<()> {
+        Err(Error::Unimplemented("checkpoint".to_string()))
+    }
+
     pub fn restore(&self) -> Result<()> {
         Err(Error::Unimplemented("restore".to_string()))
     }
 
-    /// Resume a container
-    pub fn resume(&self, id: &str) -> Result<()> {
-        let args = ["pause".to_string(), id.to_string()];
-        let _ = self.launch(self.command(&args)?, true)?;
-        Ok(())
+    /// List all the processes inside the container, returning their pids
+    pub fn ps(&self, id: &str) -> Result<Vec<usize>> {
+        let args = [
+            "ps".to_string(),
+            "--format=json".to_string(),
+            id.to_string(),
+        ];
+        let res = self.launch(self.command(&args)?, false)?;
+        let output = res.output.trim();
+
+        // Ugly hack to work around golang
+        Ok(if output == "null" {
+            Vec::new()
+        } else {
+            serde_json::from_str(output).map_err(Error::JsonDeserializationFailed)?
+        })
     }
 
     /// Run the create, start, delete lifecycle of the container and return its exit status
@@ -361,17 +282,20 @@ impl Client {
     where
         P: AsRef<Path>,
     {
-        let mut args = vec!["run".to_string(), "--bundle".to_string()];
+        let mut args = vec![
+            "run".to_string(),
+            "--bundle".to_string(),
+            utils::abs_string(bundle)?,
+        ];
         if let Some(opts) = opts {
             args.append(&mut opts.args()?);
         }
-        args.push(utils::abs_string(bundle)?);
         args.push(id.to_string());
         let mut cmd = self.command(&args)?;
-        if let Some(CreateOpts { io: Some(_io), .. }) = opts {
-            _io.set(&mut cmd).map_err(Error::UnavailableIO)?;
+        if let Some(CreateOpts { io: Some(io), .. }) = opts {
+            io.set(&mut cmd).map_err(|e| Error::IoSet(e.to_string()))?;
         };
-        self.launch(self.command(&args)?, true)
+        self.launch(cmd, true)
     }
 
     /// Start an already created container
@@ -387,12 +311,22 @@ impl Client {
         serde_json::from_str(&res.output).map_err(Error::JsonDeserializationFailed)
     }
 
+    /// Return the latest statistics for a container
+    pub fn stats(&self, id: &str) -> Result<events::Stats> {
+        let args = vec!["events".to_string(), "--stats".to_string(), id.to_string()];
+        let res = self.launch(self.command(&args)?, true)?;
+        let event: events::Event =
+            serde_json::from_str(&res.output).map_err(Error::JsonDeserializationFailed)?;
+        if let Some(stats) = event.stats {
+            Ok(stats)
+        } else {
+            Err(Error::MissingContainerStats)
+        }
+    }
+
     /// Update a container with the provided resource spec
-    pub fn update(&self, id: &str, resources: &Linux) -> Result<()> {
-        let filename = utils::temp_filename_in_runtime_dir()?;
-        let spec_json =
-            serde_json::to_string(resources).map_err(Error::JsonDeserializationFailed)?;
-        std::fs::write(&filename, spec_json).map_err(Error::SpecFileCreationFailed)?;
+    pub fn update(&self, id: &str, resources: &LinuxResources) -> Result<()> {
+        let (_temp_file, filename) = write_value_to_temp_file(resources)?;
         let args = [
             "update".to_string(),
             "--resources".to_string(),
@@ -404,68 +338,54 @@ impl Client {
     }
 }
 
-#[derive(Debug, Clone)]
-pub struct AsyncClient(runc::Runc);
+// a macro tool to cleanup the file with name $filename,
+// there is no async drop in async rust, so we have to call remove_file everytime
+// after a temp file created, before return of a function.
+// with this macro we don't have to write the match case codes everytime.
+#[cfg(feature = "async")]
+macro_rules! tc {
+    ($b:expr, $filename: expr) => {
+        match $b {
+            Ok(r) => r,
+            Err(e) => {
+                let _ = tokio::fs::remove_file($filename).await;
+                return Err(e);
+            }
+        }
+    };
+}
 
-// As monitor instance never have to be mutable (it has only &self methods), declare it as const.
-const MONITOR: DefaultMonitor = DefaultMonitor::new();
+#[cfg(not(feature = "async"))]
+pub trait Spawner: Debug {
+    fn execute(&self, cmd: Command) -> Result<(ExitStatus, u32, String, String)>;
+}
 
-/// Async client for runc
+#[cfg(feature = "async")]
+#[async_trait]
+pub trait Spawner: Debug {
+    async fn execute(&self, cmd: Command) -> Result<(ExitStatus, u32, String, String)>;
+}
+
+/// Async implementation for [Runc].
+///
 /// Note that you MUST use this client on tokio runtime, as this client internally use [`tokio::process::Command`]
 /// and some other utilities.
-impl AsyncClient {
-    /// Create a new runc client from the supplied configuration
-    pub fn from_config(mut config: Config) -> Result<Self> {
-        config.build_async()
-    }
-
-    pub fn command(&self, args: &[String]) -> Result<tokio::process::Command> {
-        let args = [&self.0.args()?, args].concat();
-        let mut cmd = tokio::process::Command::new(&self.0.command);
-        cmd.stdin(Stdio::null())
-            .stdout(Stdio::inherit())
-            .stderr(Stdio::inherit());
-        cmd.args(&args).env_remove("NOTIFY_SOCKET"); // NOTIFY_SOCKET introduces a special behavior in runc but should only be set if invoked from systemd
-        Ok(cmd)
-    }
-
-    pub async fn launch(
-        &self,
-        cmd: tokio::process::Command,
-        combined_output: bool,
-    ) -> Result<Response> {
-        let (tx, rx) = tokio::sync::oneshot::channel::<Exit>();
-        let start = MONITOR.start(cmd, tx);
-        let wait = MONITOR.wait(rx);
-        let (
-            Output {
-                status,
-                stdout,
-                stderr,
-            },
-            Exit { pid, .. },
-        ) = tokio::try_join!(start, wait).map_err(Error::InvalidCommand)?;
-
-        // ugly hack to work around
-        let stdout = String::from_utf8(stdout)
-            .expect("returned non-utf8 characters from container process.");
-        let stderr = String::from_utf8(stderr)
-            .expect("returned non-utf8 characters from container process.");
-
+#[cfg(feature = "async")]
+impl Runc {
+    async fn launch(&self, cmd: Command, combined_output: bool) -> Result<Response> {
+        debug!("Execute command {:?}", cmd);
+        let (status, pid, stdout, stderr) = self.spawner.execute(cmd).await?;
         if status.success() {
-            if combined_output {
-                Ok(Response {
-                    pid,
-                    status,
-                    output: stdout + stderr.as_str(),
-                })
+            let output = if combined_output {
+                stdout + stderr.as_str()
             } else {
-                Ok(Response {
-                    pid,
-                    status,
-                    output: stdout,
-                })
-            }
+                stdout
+            };
+            Ok(Response {
+                pid,
+                status,
+                output,
+            })
         } else {
             Err(Error::CommandFailed {
                 status,
@@ -475,12 +395,13 @@ impl AsyncClient {
         }
     }
 
-    pub async fn checkpoint(&self) -> Result<()> {
-        Err(Error::Unimplemented("checkpoint".to_string()))
-    }
-
     /// Create a new container
-    pub async fn create<P>(&self, id: &str, bundle: P, opts: Option<&CreateOpts>) -> Result<()>
+    pub async fn create<P>(
+        &self,
+        id: &str,
+        bundle: P,
+        opts: Option<&CreateOpts>,
+    ) -> Result<Response>
     where
         P: AsRef<Path>,
     {
@@ -495,36 +416,14 @@ impl AsyncClient {
         args.push(id.to_string());
         let mut cmd = self.command(&args)?;
         match opts {
-            Some(CreateOpts { io: Some(_io), .. }) => {
-                _io.set_tk(&mut cmd).map_err(Error::UnavailableIO)?;
-                let (tx, rx) = tokio::sync::oneshot::channel::<Exit>();
-                let start = MONITOR.start(cmd, tx);
-                let wait = MONITOR.wait(rx);
-                let (
-                    Output {
-                        status,
-                        stdout,
-                        stderr,
-                    },
-                    _,
-                ) = tokio::try_join!(start, wait).map_err(Error::InvalidCommand)?;
-                _io.close_after_start();
-
-                let stdout = String::from_utf8(stdout).unwrap();
-                let stderr = String::from_utf8(stderr).unwrap();
-                if !status.success() {
-                    return Err(Error::CommandFailed {
-                        status,
-                        stdout,
-                        stderr,
-                    });
-                }
-            }
-            _ => {
-                let _ = self.launch(cmd, true).await?;
+            Some(CreateOpts { io: Some(io), .. }) => {
+                io.set(&mut cmd).map_err(Error::UnavailableIO)?;
+                let res = self.launch(cmd, true).await?;
+                io.close_after_start();
+                Ok(res)
             }
+            _ => self.launch(cmd, true).await,
         }
-        Ok(())
     }
 
     /// Delete a container
@@ -539,13 +438,34 @@ impl AsyncClient {
     }
 
     /// Return an event stream of container notifications
-    pub async fn events(&self, _id: &str, _interval: &Duration) -> Result<()> {
+    pub async fn events(&self, _id: &str, _interval: &std::time::Duration) -> Result<()> {
         Err(Error::Unimplemented("events".to_string()))
     }
 
     /// Execute an additional process inside the container
-    pub async fn exec(&self, _id: &str, _spec: &Process, _opts: Option<&ExecOpts>) -> Result<()> {
-        Err(Error::Unimplemented("exec".to_string()))
+    pub async fn exec(&self, id: &str, spec: &Process, opts: Option<&ExecOpts>) -> Result<()> {
+        let f = write_value_to_temp_file(spec).await?;
+        let mut args = vec!["exec".to_string(), "--process".to_string(), f.clone()];
+        if let Some(opts) = opts {
+            args.append(&mut tc!(opts.args(), &f));
+        }
+        args.push(id.to_string());
+        let mut cmd = self.command(&args)?;
+        match opts {
+            Some(ExecOpts { io: Some(io), .. }) => {
+                tc!(
+                    io.set(&mut cmd).map_err(|e| Error::IoSet(e.to_string())),
+                    &f
+                );
+                tc!(self.launch(cmd, true).await, &f);
+                io.close_after_start();
+            }
+            _ => {
+                tc!(self.launch(cmd, true).await, &f);
+            }
+        }
+        let _ = tokio::fs::remove_file(&f).await;
+        Ok(())
     }
 
     /// Send the specified signal to processes inside the container
@@ -562,9 +482,10 @@ impl AsyncClient {
 
     /// List all containers associated with this runc instance
     pub async fn list(&self) -> Result<Vec<Container>> {
-        let args = ["list".to_string(), "--format-json".to_string()];
+        let args = ["list".to_string(), "--format=json".to_string()];
         let res = self.launch(self.command(&args)?, true).await?;
         let output = res.output.trim();
+
         // Ugly hack to work around golang
         Ok(if output == "null" {
             Vec::new()
@@ -580,15 +501,31 @@ impl AsyncClient {
         Ok(())
     }
 
+    /// Resume a container
+    pub async fn resume(&self, id: &str) -> Result<()> {
+        let args = ["resume".to_string(), id.to_string()];
+        let _ = self.launch(self.command(&args)?, true).await?;
+        Ok(())
+    }
+
+    pub async fn checkpoint(&self) -> Result<()> {
+        Err(Error::Unimplemented("checkpoint".to_string()))
+    }
+
+    pub async fn restore(&self) -> Result<()> {
+        Err(Error::Unimplemented("restore".to_string()))
+    }
+
     /// List all the processes inside the container, returning their pids
     pub async fn ps(&self, id: &str) -> Result<Vec<usize>> {
         let args = [
             "ps".to_string(),
-            "--format-json".to_string(),
+            "--format=json".to_string(),
             id.to_string(),
         ];
         let res = self.launch(self.command(&args)?, true).await?;
         let output = res.output.trim();
+
         // Ugly hack to work around golang
         Ok(if output == "null" {
             Vec::new()
@@ -597,29 +534,25 @@ impl AsyncClient {
         })
     }
 
-    pub async fn restore(&self) -> Result<()> {
-        Err(Error::Unimplemented("restore".to_string()))
-    }
-
-    /// Resume a container
-    pub async fn resume(&self, id: &str) -> Result<()> {
-        let args = ["pause".to_string(), id.to_string()];
-        let _ = self.launch(self.command(&args)?, true).await?;
-        Ok(())
-    }
-
     /// Run the create, start, delete lifecycle of the container and return its exit status
     pub async fn run<P>(&self, id: &str, bundle: P, opts: Option<&CreateOpts>) -> Result<()>
     where
         P: AsRef<Path>,
     {
-        let mut args = vec!["run".to_string(), "--bundle".to_string()];
+        let mut args = vec![
+            "run".to_string(),
+            "--bundle".to_string(),
+            utils::abs_string(bundle)?,
+        ];
         if let Some(opts) = opts {
             args.append(&mut opts.args()?);
         }
-        args.push(utils::abs_string(bundle)?);
         args.push(id.to_string());
-        let _ = self.launch(self.command(&args)?, true).await?;
+        let mut cmd = self.command(&args)?;
+        if let Some(CreateOpts { io: Some(io), .. }) = opts {
+            io.set(&mut cmd).map_err(|e| Error::IoSet(e.to_string()))?;
+        };
+        let _ = self.launch(cmd, true).await?;
         Ok(())
     }
 
@@ -638,10 +571,10 @@ impl AsyncClient {
     }
 
     /// Return the latest statistics for a container
-    pub async fn stats(&self, id: &str) -> Result<Stats> {
+    pub async fn stats(&self, id: &str) -> Result<events::Stats> {
         let args = vec!["events".to_string(), "--stats".to_string(), id.to_string()];
         let res = self.launch(self.command(&args)?, true).await?;
-        let event: Event =
+        let event: events::Event =
             serde_json::from_str(&res.output).map_err(Error::JsonDeserializationFailed)?;
         if let Some(stats) = event.stats {
             Ok(stats)
@@ -651,56 +584,46 @@ impl AsyncClient {
     }
 
     /// Update a container with the provided resource spec
-    pub async fn update(&self, id: &str, resources: &Linux) -> Result<()> {
-        let filename = utils::temp_filename_in_runtime_dir()?;
-        let spec_json =
-            serde_json::to_string(resources).map_err(Error::JsonDeserializationFailed)?;
-        std::fs::write(&filename, spec_json).map_err(Error::SpecFileCreationFailed)?;
-        let args = vec![
+    pub async fn update(&self, id: &str, resources: &LinuxResources) -> Result<()> {
+        let f = write_value_to_temp_file(resources).await?;
+        let args = [
             "update".to_string(),
             "--resources".to_string(),
-            filename,
+            f.to_string(),
             id.to_string(),
         ];
-        let _ = self.launch(self.command(&args)?, true).await?;
+        let _ = tc!(self.launch(self.command(&args)?, true).await, &f);
+        let _ = tokio::fs::remove_file(&f).await;
         Ok(())
     }
 }
 
 #[cfg(test)]
-#[cfg(target_os = "linux")]
+#[cfg(all(target_os = "linux", not(feature = "async")))]
 mod tests {
+    use std::sync::Arc;
+
+    use super::io::{InheritedStdIo, PipedStdIo};
     use super::*;
 
-    // following style of go-runc, use only true/false to test
-    const CMD_TRUE: &str = "/bin/true";
-    const CMD_FALSE: &str = "/bin/false";
-
-    fn ok_client() -> Client {
-        Config::new()
-            .command(CMD_TRUE)
+    fn ok_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/true")
             .build()
             .expect("unable to create runc instance")
     }
 
-    fn fail_client() -> Client {
-        Config::new()
-            .command(CMD_FALSE)
+    fn fail_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/false")
             .build()
             .expect("unable to create runc instance")
     }
 
-    fn ok_async_client() -> AsyncClient {
-        Config::new()
-            .command(CMD_TRUE)
-            .build_async()
-            .expect("unable to create runc instance")
-    }
-
-    fn fail_async_client() -> AsyncClient {
-        Config::new()
-            .command(CMD_FALSE)
-            .build_async()
+    fn echo_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/echo")
+            .build()
             .expect("unable to create runc instance")
     }
 
@@ -722,10 +645,13 @@ mod tests {
     fn test_create() {
         let opts = CreateOpts::new();
         let ok_runc = ok_client();
-        ok_runc
+        let response = ok_runc
             .create("fake-id", "fake-bundle", Some(&opts))
             .expect("true failed.");
-        eprintln!("ok_runc succeeded.");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(response.output.is_empty());
+
         let fail_runc = fail_client();
         match fail_runc.create("fake-id", "fake-bundle", Some(&opts)) {
             Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -748,10 +674,13 @@ mod tests {
     fn test_run() {
         let opts = CreateOpts::new();
         let ok_runc = ok_client();
-        ok_runc
+        let response = ok_runc
             .run("fake-id", "fake-bundle", Some(&opts))
             .expect("true failed.");
-        eprintln!("ok_runc succeeded.");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(response.output.is_empty());
+
         let fail_runc = fail_client();
         match fail_runc.run("fake-id", "fake-bundle", Some(&opts)) {
             Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -779,6 +708,7 @@ mod tests {
             .exec("fake-id", &proc, Some(&opts))
             .expect("true failed.");
         eprintln!("ok_runc succeeded.");
+
         let fail_runc = fail_client();
         match fail_runc.exec("fake-id", &proc, Some(&opts)) {
             Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -805,6 +735,7 @@ mod tests {
             .delete("fake-id", Some(&opts))
             .expect("true failed.");
         eprintln!("ok_runc succeeded.");
+
         let fail_runc = fail_client();
         match fail_runc.delete("fake-id", Some(&opts)) {
             Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -823,20 +754,78 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_output() {
+        // test create cmd with inherit Io, expect empty cmd output
+        let mut opts = CreateOpts::new();
+        opts.io = Some(Arc::new(InheritedStdIo::new().unwrap()));
+        let echo_runc = echo_client();
+        let response = echo_runc
+            .create("fake-id", "fake-bundle", Some(&opts))
+            .expect("echo failed.");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(response.output.is_empty());
+
+        // test create cmd with pipe Io, expect nonempty cmd output
+        let mut opts = CreateOpts::new();
+        opts.io = Some(Arc::new(PipedStdIo::new().unwrap()));
+        let echo_runc = echo_client();
+        let response = echo_runc
+            .create("fake-id", "fake-bundle", Some(&opts))
+            .expect("echo failed.");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(!response.output.is_empty());
+    }
+}
+
+/// Tokio tests
+#[cfg(test)]
+#[cfg(all(target_os = "linux", feature = "async"))]
+mod tests {
+    use std::sync::Arc;
+
+    use super::io::{InheritedStdIo, PipedStdIo};
+    use super::*;
+
+    fn ok_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/true")
+            .build()
+            .expect("unable to create runc instance")
+    }
+
+    fn fail_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/false")
+            .build()
+            .expect("unable to create runc instance")
+    }
+
+    fn echo_client() -> Runc {
+        GlobalOpts::new()
+            .command("/bin/echo")
+            .build()
+            .expect("unable to create runc instance")
+    }
+
     #[tokio::test]
     async fn test_async_create() {
         let opts = CreateOpts::new();
-        let ok_runc = ok_async_client();
+        let ok_runc = ok_client();
         let ok_task = tokio::spawn(async move {
-            ok_runc
+            let response = ok_runc
                 .create("fake-id", "fake-bundle", Some(&opts))
                 .await
                 .expect("true failed.");
-            eprintln!("ok_runc succeeded.");
+            assert_ne!(response.pid, 0);
+            assert!(response.status.success());
+            assert!(response.output.is_empty());
         });
 
         let opts = CreateOpts::new();
-        let fail_runc = fail_async_client();
+        let fail_runc = fail_client();
         let fail_task = tokio::spawn(async move {
             match fail_runc
                 .create("fake-id", "fake-bundle", Some(&opts))
@@ -864,13 +853,13 @@ mod tests {
 
     #[tokio::test]
     async fn test_async_start() {
-        let ok_runc = ok_async_client();
+        let ok_runc = ok_client();
         let ok_task = tokio::spawn(async move {
             ok_runc.start("fake-id").await.expect("true failed.");
             eprintln!("ok_runc succeeded.");
         });
 
-        let fail_runc = fail_async_client();
+        let fail_runc = fail_client();
         let fail_task = tokio::spawn(async move {
             match fail_runc.start("fake-id").await {
                 Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -896,10 +885,7 @@ mod tests {
     #[tokio::test]
     async fn test_async_run() {
         let opts = CreateOpts::new();
-        let ok_runc = Config::new()
-            .command(CMD_TRUE)
-            .build_async()
-            .expect("unable to create runc instance");
+        let ok_runc = ok_client();
         tokio::spawn(async move {
             ok_runc
                 .create("fake-id", "fake-bundle", Some(&opts))
@@ -909,10 +895,7 @@ mod tests {
         });
 
         let opts = CreateOpts::new();
-        let fail_runc = Config::new()
-            .command(CMD_FALSE)
-            .build_async()
-            .expect("unable to create runc instance");
+        let fail_runc = fail_client();
         tokio::spawn(async move {
             match fail_runc
                 .create("fake-id", "fake-bundle", Some(&opts))
@@ -940,10 +923,7 @@ mod tests {
     #[tokio::test]
     async fn test_async_delete() {
         let opts = DeleteOpts::new();
-        let ok_runc = Config::new()
-            .command(CMD_TRUE)
-            .build_async()
-            .expect("unable to create runc instance");
+        let ok_runc = ok_client();
         tokio::spawn(async move {
             ok_runc
                 .delete("fake-id", Some(&opts))
@@ -953,10 +933,7 @@ mod tests {
         });
 
         let opts = DeleteOpts::new();
-        let fail_runc = Config::new()
-            .command(CMD_FALSE)
-            .build_async()
-            .expect("unable to create runc instance");
+        let fail_runc = fail_client();
         tokio::spawn(async move {
             match fail_runc.delete("fake-id", Some(&opts)).await {
                 Ok(_) => panic!("fail_runc returned exit status 0."),
@@ -977,4 +954,65 @@ mod tests {
         .await
         .expect("tokio spawn falied.");
     }
+
+    #[tokio::test]
+    async fn test_async_output() {
+        // test create cmd with inherit Io, expect empty cmd output
+        let mut opts = CreateOpts::new();
+        opts.io = Some(Arc::new(InheritedStdIo::new().unwrap()));
+        let echo_runc = echo_client();
+        let response = echo_runc
+            .create("fake-id", "fake-bundle", Some(&opts))
+            .await
+            .expect("echo failed:");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(response.output.is_empty());
+
+        // test create cmd with pipe Io, expect nonempty cmd output
+        let mut opts = CreateOpts::new();
+        opts.io = Some(Arc::new(PipedStdIo::new().unwrap()));
+        let response = echo_runc
+            .create("fake-id", "fake-bundle", Some(&opts))
+            .await
+            .expect("echo failed:");
+        assert_ne!(response.pid, 0);
+        assert!(response.status.success());
+        assert!(!response.output.is_empty());
+    }
+}
+
+#[derive(Debug)]
+pub struct DefaultExecutor {}
+
+#[cfg(feature = "async")]
+#[async_trait]
+impl Spawner for DefaultExecutor {
+    async fn execute(&self, cmd: Command) -> Result<(ExitStatus, u32, String, String)> {
+        let mut cmd = cmd;
+        let child = cmd.spawn().map_err(Error::ProcessSpawnFailed)?;
+        let pid = child.id().unwrap();
+        let result = child
+            .wait_with_output()
+            .await
+            .map_err(Error::InvalidCommand)?;
+        let status = result.status;
+        let stdout = String::from_utf8_lossy(&result.stdout).to_string();
+        let stderr = String::from_utf8_lossy(&result.stderr).to_string();
+        Ok((status, pid, stdout, stderr))
+    }
+}
+
+#[cfg(not(feature = "async"))]
+impl Spawner for DefaultExecutor {
+    fn execute(&self, cmd: Command) -> Result<(ExitStatus, u32, String, String)> {
+        let mut cmd = cmd;
+        let child = cmd.spawn().map_err(Error::ProcessSpawnFailed)?;
+        let pid = child.id();
+        let result = child.wait_with_output().map_err(Error::InvalidCommand)?;
+        let status = result.status;
+        let stdout = String::from_utf8_lossy(&result.stdout).to_string();
+        let stderr = String::from_utf8_lossy(&result.stderr).to_string();
+        Ok((status, pid, stdout, stderr))
+    }
 }
